@@ -1,5 +1,5 @@
 /*
-    Copyright 2006,2007,2008,2009,2010 Martin Pärtel <martin.partel@gmail.com>
+    Copyright 2006,2007,2008,2009,2010,2012 Martin Pärtel <martin.partel@gmail.com>
 
     This file is part of bindfs.
 
@@ -31,7 +31,7 @@
 
 #include <config.h>
 
-/* For pread/pwrite */
+/* For pread/pwrite and readdir_r */
 #define _XOPEN_SOURCE 500
 
 #include <stdlib.h>
@@ -64,10 +64,11 @@
 #include "debug.h"
 #include "permchain.h"
 #include "userinfo.h"
+#include "usermap.h"
 #include "misc.h"
 
 /* SETTINGS */
-static struct settings {
+static struct Settings {
     const char *progname;
     struct permchain *permchain; /* permission bit rules. see permchain.h */
     uid_t new_uid; /* user-specified uid */
@@ -77,6 +78,12 @@ static struct settings {
     const char *mntsrc;
     const char *mntdest;
     int mntsrc_fd;
+    
+    char* original_working_dir;
+    mode_t original_umask;
+    
+    UserMap* usermap; /* From the --map option. */
+    UserMap* usermap_reverse;
 
     enum CreatePolicy {
         CREATE_AS_USER,
@@ -117,7 +124,11 @@ static struct settings {
     gid_t *mirrored_members;
     int num_mirrored_members;
 
+    int realistic_permissions;
+
     int ctime_from_mtime;
+    int hide_hard_links;
+    
 } settings;
 
 
@@ -132,9 +143,12 @@ static int is_mirrored_user(uid_t uid);
 /* Processes the virtual path to a real path. Don't free() the result. */
 static const char *process_path(const char *path);
 
-/* The common parts of getattr and fgetattr */
+/* The common parts of getattr and fgetattr. */
 static int getattr_common(const char *path, struct stat *stbuf);
 
+/* Chowns a new file if necessary. */
+static void chown_new_file(const char *path, struct fuse_context *fc, int (*chown_func)(const char*, uid_t, gid_t));
+
 
 /* FUSE callbacks */
 static void *bindfs_init();
@@ -174,9 +188,13 @@ static int bindfs_fsync(const char *path, int isdatasync,
 
 
 static void print_usage(const char *progname);
-static void atexit_func();
 static int process_option(void *data, const char *arg, int key,
                           struct fuse_args *outargs);
+static int parse_mirrored_users(char* mirror);
+static int parse_user_map(UserMap *map, UserMap *reverse_map, char *spec);
+static char* get_working_dir();
+static void maybe_stdout_stderr_to_file();
+static void atexit_func();
 
 static int is_mirroring_enabled()
 {
@@ -187,18 +205,14 @@ static int is_mirrored_user(uid_t uid)
 {
     int i;
     for (i = 0; i < settings.num_mirrored_users; ++i) {
-        if (settings.mirrored_users[i] == uid)
-            break;
-    }
-    if (i < settings.num_mirrored_users) { /* found in mirrored_users */
-        return 1;
-    } else {
-        for (i = 0; i < settings.num_mirrored_members; ++i) {
-            if (user_belongs_to_group(uid, settings.mirrored_members[i]))
-                break;
-        }
-        if (i < settings.num_mirrored_members) /* found in mirrored_members */
+        if (settings.mirrored_users[i] == uid) {
             return 1;
+        }
+    }
+    for (i = 0; i < settings.num_mirrored_members; ++i) {
+        if (user_belongs_to_group(uid, settings.mirrored_members[i])) {
+            return 1;
+        }
     }
     return 0;
 }
@@ -228,6 +242,10 @@ static int getattr_common(const char *procpath, struct stat *stbuf)
     if (settings.ctime_from_mtime)
         stbuf->st_ctime = stbuf->st_mtime;
 
+    /* Possibly map user/group */
+    stbuf->st_uid = usermap_get_uid_or_default(settings.usermap, stbuf->st_uid, stbuf->st_uid);
+    stbuf->st_gid = usermap_get_gid_or_default(settings.usermap, stbuf->st_gid, stbuf->st_gid);
+    
     /* Report user-defined owner/group if specified */
     if (settings.new_uid != -1)
         stbuf->st_uid = settings.new_uid;
@@ -242,27 +260,63 @@ static int getattr_common(const char *procpath, struct stat *stbuf)
         return 0;
     }
 
-    if ((stbuf->st_mode & S_IFLNK) == S_IFLNK)
-        return 0; /* don't bother with symlink permissions -- they don't matter */
+    /* Hide hard links */
+    if (settings.hide_hard_links)
+        stbuf->st_nlink = 1;
 
-    /* Apply user-defined permission bit modifications */
-    stbuf->st_mode = permchain_apply(settings.permchain, stbuf->st_mode);
+    /* Then permission bits. Symlink permissions don't matter, though. */
+    if ((stbuf->st_mode & S_IFLNK) != S_IFLNK) {
+        /* Apply user-defined permission bit modifications */
+        stbuf->st_mode = permchain_apply(settings.permchain, stbuf->st_mode);
 
-    /* Check that we can really do what we promise */
-    if (access(procpath, R_OK) == -1)
-        stbuf->st_mode &= ~0444;
-    if (access(procpath, W_OK) == -1)
-        stbuf->st_mode &= ~0222;
-    if (access(procpath, X_OK) == -1)
-        stbuf->st_mode &= ~0111;
+        /* Check that we can really do what we promise if --realistic-permissions was given */
+        if (settings.realistic_permissions) {
+            if (access(procpath, R_OK) == -1)
+                stbuf->st_mode &= ~0444;
+            if (access(procpath, W_OK) == -1)
+                stbuf->st_mode &= ~0222;
+            if (access(procpath, X_OK) == -1)
+                stbuf->st_mode &= ~0111;
+        }
+    }
 
     return 0;
 }
 
+static void chown_new_file(const char *path, struct fuse_context *fc, int (*chown_func)(const char*, uid_t, gid_t))
+{
+    uid_t file_owner;
+    gid_t file_group;
+    
+    if (settings.create_policy == CREATE_AS_USER) {
+        file_owner = fc->uid;
+        file_group = fc->gid;
+    } else {
+        file_owner = -1;
+        file_group = -1;
+    }
+    
+    file_owner = usermap_get_uid_or_default(settings.usermap_reverse, fc->uid, file_owner);
+    file_group = usermap_get_gid_or_default(settings.usermap_reverse, fc->gid, file_group);
+
+    if (settings.create_for_uid != -1)
+        file_owner = settings.create_for_uid;
+    if (settings.create_for_gid != -1)
+        file_group = settings.create_for_gid;
+
+    if ((file_owner != -1) || (file_group != -1)) {
+        if (chown_func(path, file_owner, file_group) == -1) {
+            DPRINTF("Failed to chown new file or directory (%d)", errno);
+        }
+    }
+}
+
 static void *bindfs_init()
 {
     assert(settings.permchain != NULL);
     assert(settings.mntsrc_fd > 0);
+    
+    maybe_stdout_stderr_to_file();
 
     if (fchdir(settings.mntsrc_fd) != 0) {
         fprintf(
@@ -341,12 +395,25 @@ static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                           off_t offset, struct fuse_file_info *fi)
 {
     DIR *dp = get_dirp(fi);
+    struct dirent *de_buf;
     struct dirent *de;
-
+    struct stat st;
+    int result = 0;
     (void) path;
+    
+    de_buf = malloc(offsetof(struct dirent, d_name) + pathconf(path, _PC_NAME_MAX) + 1);
+    
     seekdir(dp, offset);
-    while ((de = readdir(dp)) != NULL) {
-        struct stat st;
+    while (1) {
+        result = readdir_r(dp, de_buf, &de);
+        if (result != 0) {
+            result = -result;
+            break;
+        }
+        if (de == NULL) {
+            break;
+        }
+        
         memset(&st, 0, sizeof(st));
         st.st_ino = de->d_ino;
         st.st_mode = de->d_type << 12;
@@ -354,7 +421,8 @@ static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
             break;
     }
 
-    return 0;
+    free(de_buf);
+    return result;
 }
 
 static int bindfs_releasedir(const char *path, struct fuse_file_info *fi)
@@ -369,8 +437,6 @@ static int bindfs_mknod(const char *path, mode_t mode, dev_t rdev)
 {
     int res;
     struct fuse_context *fc;
-    uid_t file_owner = -1;
-    gid_t file_group = -1;
 
     path = process_path(path);
 
@@ -383,22 +449,8 @@ static int bindfs_mknod(const char *path, mode_t mode, dev_t rdev)
     if (res == -1)
         return -errno;
 
-    if (settings.create_policy == CREATE_AS_USER) {
-        fc = fuse_get_context();
-        file_owner = fc->uid;
-        file_group = fc->gid;
-    }
-
-    if (settings.create_for_uid != -1)
-        file_owner = settings.create_for_uid;
-    if (settings.create_for_gid != -1)
-        file_group = settings.create_for_gid;
-
-    if ((file_owner != -1) || (file_group != -1)) {
-        if (chown(path, file_owner, file_group) == -1) {
-            DPRINTF("Failed to chown new device node (%d)", errno);
-        }
-    }
+    fc = fuse_get_context();
+    chown_new_file(path, fc, &chown);
 
     return 0;
 }
@@ -407,8 +459,6 @@ static int bindfs_mkdir(const char *path, mode_t mode)
 {
     int res;
     struct fuse_context *fc;
-    uid_t file_owner = -1;
-    gid_t file_group = -1;
 
     path = process_path(path);
 
@@ -419,22 +469,8 @@ static int bindfs_mkdir(const char *path, mode_t mode)
     if (res == -1)
         return -errno;
 
-    if (settings.create_policy == CREATE_AS_USER) {
-        fc = fuse_get_context();
-        file_owner = fc->uid;
-        file_group = fc->gid;
-    }
-
-    if (settings.create_for_uid != -1)
-        file_owner = settings.create_for_uid;
-    if (settings.create_for_gid != -1)
-        file_group = settings.create_for_gid;
-
-    if ((file_owner != -1) || (file_group != -1)) {
-        if (chown(path, file_owner, file_group) == -1) {
-            DPRINTF("Failed to chown new directory (%d)", errno);
-        }
-    }
+    fc = fuse_get_context();
+    chown_new_file(path, fc, &chown);
 
     return 0;
 }
@@ -469,8 +505,6 @@ static int bindfs_symlink(const char *from, const char *to)
 {
     int res;
     struct fuse_context *fc;
-    uid_t file_owner = -1;
-    gid_t file_group = -1;
 
     to = process_path(to);
 
@@ -478,22 +512,8 @@ static int bindfs_symlink(const char *from, const char *to)
     if (res == -1)
         return -errno;
 
-    if (settings.create_policy == CREATE_AS_USER) {
-        fc = fuse_get_context();
-        file_owner = fc->uid;
-        file_group = fc->gid;
-    }
-
-    if (settings.create_for_uid != -1)
-        file_owner = settings.create_for_uid;
-    if (settings.create_for_gid != -1)
-        file_group = settings.create_for_gid;
-
-    if ((file_owner != -1) || (file_group != -1)) {
-        if (lchown(to, file_owner, file_group) == -1) {
-            DPRINTF("Failed to lchown new symlink (%d)", errno);
-        }
-    }
+    fc = fuse_get_context();
+    chown_new_file(to, fc, &lchown);
 
     return 0;
 }
@@ -580,6 +600,7 @@ static int bindfs_chown(const char *path, uid_t uid, gid_t gid)
     if (uid != -1) {
         switch (settings.chown_policy) {
         case CHOWN_NORMAL:
+            uid = usermap_get_uid_or_default(settings.usermap_reverse, uid, uid);
             break;
         case CHOWN_IGNORE:
             uid = -1;
@@ -592,6 +613,7 @@ static int bindfs_chown(const char *path, uid_t uid, gid_t gid)
     if (gid != -1) {
         switch (settings.chgrp_policy) {
         case CHGRP_NORMAL:
+            gid = usermap_get_gid_or_default(settings.usermap_reverse, gid, gid);
             break;
         case CHGRP_IGNORE:
             gid = -1;
@@ -628,7 +650,6 @@ static int bindfs_ftruncate(const char *path, off_t size,
                             struct fuse_file_info *fi)
 {
     int res;
-
     (void) path;
 
     res = ftruncate(fi->fh, size);
@@ -655,8 +676,6 @@ static int bindfs_create(const char *path, mode_t mode, struct fuse_file_info *f
 {
     int fd;
     struct fuse_context *fc;
-    uid_t file_owner = -1;
-    gid_t file_group = -1;
 
     path = process_path(path);
 
@@ -667,22 +686,8 @@ static int bindfs_create(const char *path, mode_t mode, struct fuse_file_info *f
     if (fd == -1)
         return -errno;
 
-    if (settings.create_policy == CREATE_AS_USER) {
-        fc = fuse_get_context();
-        file_owner = fc->uid;
-        file_group = fc->gid;
-    }
-
-    if (settings.create_for_uid != -1)
-        file_owner = settings.create_for_uid;
-    if (settings.create_for_gid != -1)
-        file_group = settings.create_for_gid;
-
-    if ((file_owner != -1) || (file_group != -1)) {
-        if (chown(path, file_owner, file_group) == -1) {
-            DPRINTF("Failed to chown new file (%d)", errno);
-        }
-    }
+    fc = fuse_get_context();
+    chown_new_file(path, fc, &chown);
 
     fi->fh = fd;
     return 0;
@@ -706,8 +711,8 @@ static int bindfs_read(const char *path, char *buf, size_t size, off_t offset,
                        struct fuse_file_info *fi)
 {
     int res;
-
     (void) path;
+    
     res = pread(fi->fh, buf, size, offset);
     if (res == -1)
         res = -errno;
@@ -719,8 +724,8 @@ static int bindfs_write(const char *path, const char *buf, size_t size,
                         off_t offset, struct fuse_file_info *fi)
 {
     int res;
-
     (void) path;
+
     res = pwrite(fi->fh, buf, size, offset);
     if (res == -1)
         res = -errno;
@@ -744,6 +749,7 @@ static int bindfs_statfs(const char *path, struct statvfs *stbuf)
 static int bindfs_release(const char *path, struct fuse_file_info *fi)
 {
     (void) path;
+    
     close(fi->fh);
 
     return 0;
@@ -896,14 +902,14 @@ static void print_usage(const char *progname)
            "  -h      --help            Print this and exit.\n"
            "  -V      --version         Print version number and exit.\n"
            "\n"
-           "Options:\n"
+           "File ownership:\n"
            "  -u      --user, --owner   Set file owner.\n"
            "  -g      --group           Set file group.\n"
            "  -m      --mirror          Comma-separated list of users who will see\n"
            "                            themselves as the owners of all files.\n"
            "  -M      --mirror-only     Like --mirror but disallow access for\n"
            "                            all other users.\n"
-           "  -n      --no-allow-other  Do not add -o allow_other to fuse options.\n"
+           " --map=user1/user2:...      Let user2 see files of user1 as his own.\n"
            "\n"
            "Permission bits:\n"
            "  -p      --perms           Specify permissions, similar to chmod\n"
@@ -937,9 +943,12 @@ static void print_usage(const char *progname)
            "  --xattr-ro                Read-only xattr operations.\n"
            "  --xattr-rw                Read-write xattr operations (the default).\n"
            "\n"
-           "Time-related:\n"
+           "Miscellaneous:\n"
+           "  -n      --no-allow-other  Do not add -o allow_other to fuse options.\n"
+           "  --realistic-permissions   Hide permission bits for actions mounter can't do.\n"
            "  --ctime-from-mtime        Read file properties' change time\n"
            "                            from file content modification time.\n"
+           "  --hide-hard-links         Always report a hard link count of 1.\n"
            "\n"
            "FUSE options:\n"
            "  -o opt[,opt,...]          Mount options.\n"
@@ -954,18 +963,6 @@ static void print_usage(const char *progname)
 }
 
 
-static void atexit_func()
-{
-    permchain_destroy(settings.permchain);
-    settings.permchain = NULL;
-    permchain_destroy(settings.create_permchain);
-    settings.create_permchain = NULL;
-    free(settings.mirrored_users);
-    settings.mirrored_users = NULL;
-    free(settings.mirrored_members);
-    settings.mirrored_members = NULL;
-}
-
 enum OptionKey {
     OPTKEY_NONOPTION = -2,
     OPTKEY_UNKNOWN = -1,
@@ -986,7 +983,9 @@ enum OptionKey {
     OPTKEY_XATTR_NONE,
     OPTKEY_XATTR_READ_ONLY,
     OPTKEY_XATTR_READ_WRITE,
-    OPTKEY_CTIME_FROM_MTIME
+    OPTKEY_REALISTIC_PERMISSIONS,
+    OPTKEY_CTIME_FROM_MTIME,
+    OPTKEY_HIDE_HARD_LINKS
 };
 
 static int process_option(void *data, const char *arg, int key,
@@ -1058,9 +1057,15 @@ static int process_option(void *data, const char *arg, int key,
         settings.xattr_policy = XATTR_READ_WRITE;
         return 0;
 
+    case OPTKEY_REALISTIC_PERMISSIONS:
+        settings.realistic_permissions = 1;
+        return 0;
     case OPTKEY_CTIME_FROM_MTIME:
         settings.ctime_from_mtime = 1;
-	return 0;
+        return 0;
+    case OPTKEY_HIDE_HARD_LINKS:
+        settings.hide_hard_links = 1;
+        return 0;
 
     case OPTKEY_NONOPTION:
         if (!settings.mntsrc) {
@@ -1079,23 +1084,213 @@ static int process_option(void *data, const char *arg, int key,
     }
 }
 
+static int parse_mirrored_users(char* mirror)
+{
+    int i;
+    int j;
+    char *p, *tmpstr;
+
+    settings.num_mirrored_users = count_chars(mirror, ',') +
+                                    count_chars(mirror, ':') + 1;
+    settings.num_mirrored_members = ((*mirror == '@') ? 1 : 0) +
+                                    count_substrs(mirror, ",@") +
+                                    count_substrs(mirror, ":@");
+    settings.num_mirrored_users -= settings.num_mirrored_members;
+    settings.mirrored_users = malloc(settings.num_mirrored_users*sizeof(uid_t));
+    settings.mirrored_members = malloc(settings.num_mirrored_members*sizeof(gid_t));
+
+    i = 0; /* iterate over mirrored_users */
+    j = 0; /* iterate over mirrored_members */
+    p = mirror;
+    while (i < settings.num_mirrored_users || j < settings.num_mirrored_members) {
+        tmpstr = strdup_until(p, ",:");
+
+        if (*tmpstr == '@') { /* This is a group name */
+            if (!group_gid(tmpstr + 1, &settings.mirrored_members[j++])) {
+                fprintf(stderr, "Invalid group ID: '%s'\n", tmpstr + 1);
+                free(tmpstr);
+                return 0;
+            }
+        } else {
+            if (!user_uid(tmpstr, &settings.mirrored_users[i++])) {
+                fprintf(stderr, "Invalid user ID: '%s'\n", tmpstr);
+                free(tmpstr);
+                return 0;
+            }
+        }
+        free(tmpstr);
+
+        while (*p != '\0' && *p != ',' && *p != ':') {
+            ++p;
+        }
+        if (*p != '\0') {
+            ++p;
+        } else {
+            /* Done. The counters should match. */
+            assert(i == settings.num_mirrored_users);
+            assert(j == settings.num_mirrored_members);
+        }
+    }
+    
+    return 1;
+}
+
+static int parse_user_map(UserMap *map, UserMap *reverse_map, char *spec)
+{
+    char *p = spec;
+    char *tmpstr = NULL;
+    char *q;
+    uid_t uid_from, uid_to;
+    gid_t gid_from, gid_to;
+    UsermapStatus status;
+    
+    while (*p != '\0') {
+        free(tmpstr);
+        tmpstr = strdup_until(p, ",:");
+        
+        if (tmpstr[0] == '@') { /* group */
+            q = strstr(tmpstr, "/@");
+            if (!q) {
+                fprintf(stderr, "Invalid syntax: expected @group1/@group2 but got `%s`\n", tmpstr);
+                goto fail;
+            }
+            *q = '\0';
+            if (!group_gid(tmpstr + 1, &gid_from)) {
+                fprintf(stderr, "Invalid group: %s\n", tmpstr);
+                goto fail;
+            }
+            q += strlen("/@");
+            if (!group_gid(q, &gid_to)) {
+                fprintf(stderr, "Invalid group: %s\n", tmpstr);
+                goto fail;
+            }
+            
+            status = usermap_add_gid(map, gid_from, gid_to);
+            if (status != 0) {
+                fprintf(stderr, "%s\n", usermap_errorstr(status));
+                goto fail;
+            }
+            status = usermap_add_gid(reverse_map, gid_to, gid_from);
+            if (status != 0) {
+                fprintf(stderr, "%s\n", usermap_errorstr(status));
+                goto fail;
+            }
+            
+        } else {
+            
+            q = strstr(tmpstr, "/");
+            if (!q) {
+                fprintf(stderr, "Invalid syntax: expected user1/user2 but got `%s`\n", tmpstr);
+                goto fail;
+            }
+            *q = '\0';
+            if (!user_uid(tmpstr, &uid_from)) {
+                fprintf(stderr, "Invalid username: %s\n", tmpstr);
+                goto fail;
+            }
+            q += strlen("/");
+            if (!user_uid(q, &uid_to)) {
+                fprintf(stderr, "Invalid username: %s\n", tmpstr);
+                goto fail;
+            }
+            
+            status = usermap_add_uid(map, uid_from, uid_to);
+            if (status != 0) {
+                fprintf(stderr, "%s\n", usermap_errorstr(status));
+                goto fail;
+            }
+            status = usermap_add_uid(reverse_map, uid_to, uid_from);
+            if (status != 0) {
+                fprintf(stderr, "%s\n", usermap_errorstr(status));
+                goto fail;
+            }
+        }
+        
+        while (*p != '\0' && *p != ',' && *p != ':') {
+            ++p;
+        }
+        if (*p != '\0') {
+            ++p;
+        }
+    }
+    
+    free(tmpstr);
+    return 1;
+    
+fail:
+    free(tmpstr);
+    return 0;
+}
+
+static void maybe_stdout_stderr_to_file()
+{
+    /* TODO: make this a command line option. */
+#if 0
+    int fd;
+    
+    const char *filename = "bindfs.log";
+    char *path = malloc(strlen(settings.original_working_dir) + 1 + strlen(filename) + 1);
+    strcpy(path, settings.original_working_dir);
+    strcat(path, "/");
+    strcat(path, filename);
+    
+    fd = open(path, O_CREAT | O_WRONLY);
+    free(path);
+    
+    fchmod(fd, 0777 & ~settings.original_umask);
+    fflush(stdout);
+    fflush(stderr);
+    dup2(fd, 1);
+    dup2(fd, 2);
+#endif
+}
+
+static char* get_working_dir()
+{
+    size_t buf_size = 4096;
+    char* buf = malloc(buf_size);
+    while (!getcwd(buf, buf_size)) {
+        buf_size *= 2;
+        buf = realloc(buf, buf_size);
+    }
+    return buf;
+}
+
+static void atexit_func()
+{
+    free(settings.original_working_dir);
+    settings.original_working_dir = NULL;
+    usermap_destroy(settings.usermap);
+    settings.usermap = NULL;
+    usermap_destroy(settings.usermap_reverse);
+    settings.usermap_reverse = NULL;
+    permchain_destroy(settings.permchain);
+    settings.permchain = NULL;
+    permchain_destroy(settings.create_permchain);
+    settings.create_permchain = NULL;
+    free(settings.mirrored_users);
+    settings.mirrored_users = NULL;
+    free(settings.mirrored_members);
+    settings.mirrored_members = NULL;
+}
 
 int main(int argc, char *argv[])
 {
     struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
 
     /* Fuse's option parser will store things here. */
-    static struct OptionData {
+    struct OptionData {
         char *user;
         char *group;
         char *perms;
         char *mirror;
         char *mirror_only;
+        char *map;
         char *create_for_user;
         char *create_for_group;
         char *create_with_perms;
         int no_allow_other;
-    } od = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 0};
+    } od;
 
     #define OPT2(one, two, key) \
             FUSE_OPT_KEY(one, key), \
@@ -1116,6 +1311,7 @@ int main(int argc, char *argv[])
         OPT_OFFSET3("-p %s", "--perms=%s", "perms=%s", perms, -1),
         OPT_OFFSET3("-m %s", "--mirror=%s", "mirror=%s", mirror, -1),
         OPT_OFFSET3("-M %s", "--mirror-only=%s", "mirror-only=%s", mirror_only, -1),
+        OPT_OFFSET2("--map=%s", "map=%s", map, -1),
         OPT_OFFSET3("-n", "--no-allow-other", "no-allow-other", no_allow_other, -1),
         OPT2("--create-as-user", "create-as-user", OPTKEY_CREATE_AS_USER),
         OPT2("--create-as-mounter", "create-as-mounter", OPTKEY_CREATE_AS_MOUNTER),
@@ -1135,26 +1331,28 @@ int main(int argc, char *argv[])
         OPT2("--xattr-none", "xattr-none", OPTKEY_XATTR_NONE),
         OPT2("--xattr-ro", "xattr-ro", OPTKEY_XATTR_READ_ONLY),
         OPT2("--xattr-rw", "xattr-rw", OPTKEY_XATTR_READ_WRITE),
+        OPT2("--realistic-permissions", "realistic-permissions", OPTKEY_REALISTIC_PERMISSIONS),
         OPT2("--ctime-from-mtime", "ctime-from-mtime", OPTKEY_CTIME_FROM_MTIME),
+        OPT2("--hide-hard-links", "hide-hard-links", OPTKEY_HIDE_HARD_LINKS),
         FUSE_OPT_END
     };
 
-    /* General-purpose variables */
-    int i, j;
-    char *p, *tmpstr;
-
     int fuse_main_return;
 
 
     /* Initialize settings */
+    memset(&od, 0, sizeof(od));
     settings.progname = argv[0];
     settings.permchain = permchain_create();
+    settings.usermap = usermap_create();
+    settings.usermap_reverse = usermap_create();
     settings.new_uid = -1;
     settings.new_gid = -1;
     settings.create_for_uid = -1;
     settings.create_for_gid = -1;
     settings.mntsrc = NULL;
     settings.mntdest = NULL;
+    settings.original_working_dir = get_working_dir();
     settings.create_policy = (getuid() == 0) ? CREATE_AS_USER : CREATE_AS_MOUNTER;
     settings.create_permchain = permchain_create();
     settings.chown_policy = CHOWN_NORMAL;
@@ -1167,9 +1365,11 @@ int main(int argc, char *argv[])
     settings.num_mirrored_users = 0;
     settings.mirrored_members = NULL;
     settings.num_mirrored_members = 0;
+    settings.realistic_permissions = 0;
     settings.ctime_from_mtime = 0;
+    settings.hide_hard_links = 0;
     atexit(&atexit_func);
-
+    
     /* Parse options */
     if (fuse_opt_parse(&args, &od, options, &process_option) == -1)
         return 1;
@@ -1193,6 +1393,18 @@ int main(int argc, char *argv[])
             return 1;
         }
     }
+    
+    /* Parse usermap */
+    if (od.map) {
+        if (getuid() != 0) {
+            fprintf(stderr, "Error: You need to be root to use --map !\n");
+            return 1;
+        }
+        if (!parse_user_map(settings.usermap, settings.usermap_reverse, od.map)) {
+            /* parse_user_map printed an error */
+            return 1;
+        }
+    }
 
     /* Parse user and group for new creates */
     if (od.create_for_user) {
@@ -1226,45 +1438,8 @@ int main(int argc, char *argv[])
         od.mirror = od.mirror_only;
     }
     if (od.mirror) {
-        settings.num_mirrored_users = count_chars(od.mirror, ',') +
-                                      count_chars(od.mirror, ':') + 1;
-        settings.num_mirrored_members = ((*od.mirror == '@') ? 1 : 0) +
-                                        count_substrs(od.mirror, ",@") +
-                                        count_substrs(od.mirror, ":@");
-        settings.num_mirrored_users -= settings.num_mirrored_members;
-        settings.mirrored_users = malloc(settings.num_mirrored_users*sizeof(uid_t));
-        settings.mirrored_members = malloc(settings.num_mirrored_members*sizeof(gid_t));
-
-        i = 0; /* iterate over mirrored_users */
-        j = 0; /* iterate over mirrored_members */
-        p = od.mirror;
-        while (i < settings.num_mirrored_users || j < settings.num_mirrored_members) {
-            tmpstr = strdup_until(p, ",:");
-
-            if (*tmpstr == '@') { /* This is a group name */
-                if (!group_gid(tmpstr + 1, &settings.mirrored_members[j++])) {
-                    fprintf(stderr, "Invalid group ID: '%s'\n", tmpstr + 1);
-                    free(tmpstr);
-                    return 1;
-                }
-            } else {
-                if (!user_uid(tmpstr, &settings.mirrored_users[i++])) {
-                    fprintf(stderr, "Invalid user ID: '%s'\n", tmpstr);
-                    free(tmpstr);
-                    return 1;
-                }
-            }
-            free(tmpstr);
-
-            while (*p != '\0' && *p != ',' && *p != ':')
-                ++p;
-            if (*p != '\0')
-                ++p;
-            else {
-                /* Done. The counters should match. */
-                assert(i == settings.num_mirrored_users);
-                assert(j == settings.num_mirrored_members);
-            }
+        if (!parse_mirrored_users(od.mirror)) {
+            return 0;
         }
     }
 
@@ -1291,6 +1466,10 @@ int main(int argc, char *argv[])
     /* We want the kernel to do our access checks for us based on what getattr gives it. */
     fuse_opt_add_arg(&args, "-odefault_permissions");
     
+    /* We want to mirror inodes. */
+    fuse_opt_add_arg(&args, "-ouse_ino");
+    fuse_opt_add_arg(&args, "-oreaddir_ino");
+    
     /* We need to disable the attribute cache whenever two users
        can see different attributes. For now, only mirroring can do that. */
     if (is_mirroring_enabled()) {
@@ -1310,7 +1489,7 @@ int main(int argc, char *argv[])
     }
 
     /* Ignore the umask of the mounter on file creation */
-    umask(0);
+    settings.original_umask = umask(0);
 
     /* Remove xattr implementation if the user doesn't want it */
     if (settings.xattr_policy == XATTR_UNIMPLEMENTED) {
