@@ -1,6 +1,6 @@
 #!/usr/bin/env ruby
 #
-#   Copyright 2006,2007,2008,2009,2010 Martin Pärtel <martin.partel@gmail.com>
+#   Copyright 2006,2007,2008,2009,2010,2012 Martin Pärtel <martin.partel@gmail.com>
 #
 #   This file is part of bindfs.
 #
@@ -18,7 +18,7 @@
 #   along with bindfs.  If not, see <http://www.gnu.org/licenses/>.
 #
 
-require 'fileutils.rb'
+require 'fileutils'
 include FileUtils
 
 # Set the default umask for all tests
@@ -30,22 +30,42 @@ TESTDIR_NAME = 'tmp_test_bindfs'
 # If set to an array of test names, only those will be run
 $only_these_tests = nil
 
+def fail(msg, error = nil, options = {})
+    options = {:exit => false}.merge(options)
+    if options[:msg]
+        $stderr.puts(options[:msg])
+    end
+    if error.is_a?(Exception)
+        $stderr.puts(error.message + "\n  " + error.backtrace.join("\n  "))
+    elsif error != nil
+        $stderr.puts(error.to_s)
+    end
+    exit! 1 if options[:exit]
+end
+
+def fail!(msg, error = nil, options = {})
+    options = {:exit => true}.merge(options)
+    fail(msg, error, options)
+end
+
 # Prepares a test environment with a mounted directory
-def testenv(bindfs_args, &block)
+def testenv(bindfs_args, options = {}, &block)
 
-    testcase_title = bindfs_args
+    options = {
+        :title => bindfs_args
+    }.merge(options)
+  
+    # todo: less repetitive and more careful error handling and cleanup
 
-    return unless $only_these_tests == nil or $only_these_tests.member? testcase_title
+    return unless $only_these_tests == nil or $only_these_tests.member? options[:title]
 
-    puts "--- #{testcase_title} ---"
+    puts "--- #{options[:title]} ---"
     puts "[  #{bindfs_args}  ]"
 
     begin
         Dir.mkdir TESTDIR_NAME
     rescue Exception => ex
-        $stderr.puts "ERROR creating testdir at #{TESTDIR_NAME}"
-        $stderr.puts ex
-        exit! 1
+        fail!("ERROR creating testdir at #{TESTDIR_NAME}", ex)
     end
 
     begin
@@ -53,9 +73,7 @@ def testenv(bindfs_args, &block)
         Dir.mkdir 'src'
         Dir.mkdir 'mnt'
     rescue Exception => ex
-        $stderr.puts "ERROR preparing testdir at #{TESTDIR_NAME}"
-        $stderr.puts ex
-        exit! 1
+        fail!("ERROR preparing testdir at #{TESTDIR_NAME}", ex)
     end
 
     bindfs_pid = nil
@@ -66,25 +84,30 @@ def testenv(bindfs_args, &block)
             exit! 127
         end
     rescue Exception => ex
-        $stderr.puts "ERROR running bindfs"
-        $stderr.puts ex
-        Dir.chdir '..'
         system("rm -Rf #{TESTDIR_NAME}")
-        exit! 1
+        fail!("ERROR running bindfs", ex)
     end
 
     # Wait for bindfs to daemonize itself
     Process.wait bindfs_pid
 
-    # TODO: check that mounting was successful
+    # fuse4bsd is sometimes slow to activate the mount.
+    # http://www.freebsd.org/cgi/query-pr.cgi?pr=159102
+    if `uname -a`.downcase.include? 'freebsd'
+        puts "Waiting for fuse4bsd"
+        sleep 0.5
+    end
+
+    # Check that mount appeared in `mount`
+    if !`mount`.include? TESTDIR_NAME
+        fail!("Mount point did not appear in `mount`")
+    end
 
     testcase_ok = true
     begin
         yield
     rescue Exception => ex
-        $stderr.puts "ERROR: testcase `#{testcase_title}' failed"
-        $stderr.puts ex
-        $stderr.puts ex.backtrace
+        fail("ERROR: testcase `#{options[:title]}' failed", ex)
         testcase_ok = false
     end
 
@@ -93,24 +116,22 @@ def testenv(bindfs_args, &block)
             raise Exception.new(umount_cmd + " failed with status #{$?}")
         end
     rescue Exception => ex
-        $stderr.puts "ERROR: failed to umount"
-        $stderr.puts ex
-        $stderr.puts ex.backtrace
+        fail("ERROR: failed to umount")
         testcase_ok = false
     end
+    
+    if File.exist?("bindfs.log")
+        system("cat bindfs.log")
+    end
 
     begin
         Dir.chdir '..'
     rescue Exception => ex
-        $stderr.puts "ERROR: failed to exit test env"
-        $stderr.puts ex
-        $stderr.puts ex.backtrace
-        exit! 1
+        fail!("ERROR: failed to exit test env")
     end
 
     unless system "rm -Rf #{TESTDIR_NAME}"
-        $stderr.puts "ERROR: failed to clear test directory"
-        exit! 1
+        fail!("ERROR: failed to clear test directory")
     end
 
     if testcase_ok
@@ -121,9 +142,9 @@ def testenv(bindfs_args, &block)
 end
 
 # Like testenv but skips the test if not running as root
-def root_testenv(bindfs_args, &block)
+def root_testenv(bindfs_args, options = {}, &block)
     if Process.uid == 0
-        testenv(bindfs_args, &block)
+        testenv(bindfs_args, options, &block)
     else
         puts "--- #{bindfs_args} ---"
         puts "[  #{bindfs_args}  ]"
